//========================================================================
//
// TextOutputDev.cc
//
// Copyright 1997 Derek B. Noonburg
//
// Changed 1999 by G.Ovtcharov
//========================================================================

#ifdef __GNUC__
#pragma implementation
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <ctype.h>
#include "GString.h"
#include "gmem.h"
#include "config.h"
#include "Error.h"
#include "GfxState.h"
#include "FontEncoding.h"
#include "HtmlOutputDev.h"
#include "Image.h"
#include "GVector.h" 
#include "HtmlFonts.h"


#include "TextOutputFontInfo.h"

#define xoutRound(x) ((int)(x + 0.5))

//------------------------------------------------------------------------
// Character substitutions
//------------------------------------------------------------------------

static char *isoLatin1Subst[] = {
  "L",                          // Lslash
  "OE",                         // OE
  "S",                          // Scaron
  "Y",                          // Ydieresis
  "Z",                          // Zcaron
  "fi", "fl",                   // ligatures
  "ff", "ffi", "ffl",           // ligatures
  "i",                          // dotlessi
  "l",                          // lslash
  "oe",                         // oe
  "s",                          // scaron
  "z",                          // zcaron
  "*",                          // bullet
  "...",                        // ellipsis
  "-", "-",                     // emdash, hyphen
  "\"", "\"",                   // quotedblleft, quotedblright
  "'",                          // quotesingle
  "TM"                          // trademark
};

static char *ascii7Subst[] = {
  "A", "A", "A", "A",           // A{acute,circumflex,dieresis,grave}
  "A", "A",                     // A{ring,tilde}
  "AE",                         // AE
  "C",                          // Ccedilla
  "E", "E", "E", "E",           // E{acute,circumflex,dieresis,grave}
  "I", "I", "I", "I",           // I{acute,circumflex,dieresis,grave}
  "L",                          // Lslash
  "N",                          // Ntilde
  "O", "O", "O", "O",           // O{acute,circumflex,dieresis,grave}
  "O", "O",                     // O{slash,tilde}
  "OE",                         // OE
  "S",                          // Scaron
  "U", "U", "U", "U",           // U{acute,circumflex,dieresis,grave}
  "Y", "Y",                     // T{acute,dieresis}
  "Z",                          // Zcaron
  "a", "a", "a", "a",           // a{acute,circumflex,dieresis,grave}
  "a", "a",                     // a{ring,tilde}
  "ae",                         // ae
  "c",                          // ccedilla
  "e", "e", "e", "e",           // e{acute,circumflex,dieresis,grave}
  "fi", "fl",                   // ligatures
  "ff", "ffi", "ffl",           // ligatures
  "i",                          // dotlessi
  "i", "i", "i", "i",           // i{acute,circumflex,dieresis,grave}
  "l",                          // lslash
  "n",                          // ntilde
  "o", "o", "o", "o",           // o{acute,circumflex,dieresis,grave}
  "o", "o",                     // o{slash,tilde}
  "oe",                         // oe
  "s",                          // scaron
  "u", "u", "u", "u",           // u{acute,circumflex,dieresis,grave}
  "y", "y",                     // t{acute,dieresis}
  "z",                          // zcaron
  "|",                          // brokenbar
  "*",                          // bullet
  "...",                        // ellipsis
  "-", "-", "-",                // emdash, endash, hyphen
  "\"", "\"",                   // quotedblleft, quotedblright
  "'",                          // quotesingle
  "(R)",                        // registered
  "TM"                          // trademark
};

//------------------------------------------------------------------------
// 16-bit fonts
//------------------------------------------------------------------------

#if JAPANESE_SUPPORT

// CID 0 .. 96
static Gushort japan12Map[96] = {
  0x2120, 0x2120, 0x212a, 0x2149, 0x2174, 0x2170, 0x2173, 0x2175, // 00 .. 07
  0x2147, 0x214a, 0x214b, 0x2176, 0x215c, 0x2124, 0x213e, 0x2123, // 08 .. 0f
  0x213f, 0x2330, 0x2331, 0x2332, 0x2333, 0x2334, 0x2335, 0x2336, // 10 .. 17
  0x2337, 0x2338, 0x2339, 0x2127, 0x2128, 0x2163, 0x2161, 0x2164, // 18 .. 1f
  0x2129, 0x2177, 0x2341, 0x2342, 0x2343, 0x2344, 0x2345, 0x2346, // 20 .. 27
  0x2347, 0x2348, 0x2349, 0x234a, 0x234b, 0x234c, 0x234d, 0x234e, // 28 .. 2f
  0x234f, 0x2350, 0x2351, 0x2352, 0x2353, 0x2354, 0x2355, 0x2356, // 30 .. 37
  0x2357, 0x2358, 0x2359, 0x235a, 0x214e, 0x216f, 0x214f, 0x2130, // 38 .. 3f
  0x2132, 0x2146, 0x2361, 0x2362, 0x2363, 0x2364, 0x2365, 0x2366, // 40 .. 47
  0x2367, 0x2368, 0x2369, 0x236a, 0x236b, 0x236c, 0x236d, 0x236e, // 48 .. 4f
  0x236f, 0x2370, 0x2371, 0x2372, 0x2373, 0x2374, 0x2375, 0x2376, // 50 .. 57
  0x2377, 0x2378, 0x2379, 0x237a, 0x2150, 0x2143, 0x2151, 0x2141  // 58 .. 5f
};

// CID 325 .. 421
static Gushort japan12KanaMap1[97] = {
  0x2131, 0x2121, 0x2123, 0x2156, 0x2157, 0x2122, 0x2126, 0x2572,
  0x2521, 0x2523, 0x2525, 0x2527, 0x2529, 0x2563, 0x2565, 0x2567,
  0x2543, 0x213c, 0x2522, 0x2524, 0x2526, 0x2528, 0x252a, 0x252b,
  0x252d, 0x252f, 0x2531, 0x2533, 0x2535, 0x2537, 0x2539, 0x253b,
  0x253d, 0x253f, 0x2541, 0x2544, 0x2546, 0x2548, 0x254a, 0x254b,
  0x254c, 0x254d, 0x254e, 0x254f, 0x2552, 0x2555, 0x2558, 0x255b,
  0x255e, 0x255f, 0x2560, 0x2561, 0x2562, 0x2564, 0x2566, 0x2568,
  0x2569, 0x256a, 0x256b, 0x256c, 0x256d, 0x256f, 0x2573, 0x212b,
  0x212c, 0x212e, 0x2570, 0x2571, 0x256e, 0x2575, 0x2576, 0x2574,
  0x252c, 0x252e, 0x2530, 0x2532, 0x2534, 0x2536, 0x2538, 0x253a,
  0x253c, 0x253e, 0x2540, 0x2542, 0x2545, 0x2547, 0x2549, 0x2550,
  0x2551, 0x2553, 0x2554, 0x2556, 0x2557, 0x2559, 0x255a, 0x255c,
  0x255d
};

// CID 501 .. 598
static Gushort japan12KanaMap2[98] = {
  0x212d, 0x212f, 0x216d, 0x214c, 0x214d, 0x2152, 0x2153, 0x2154,
  0x2155, 0x2158, 0x2159, 0x215a, 0x215b, 0x213d, 0x2121, 0x2472,
  0x2421, 0x2423, 0x2425, 0x2427, 0x2429, 0x2463, 0x2465, 0x2467,
  0x2443, 0x2422, 0x2424, 0x2426, 0x2428, 0x242a, 0x242b, 0x242d,
  0x242f, 0x2431, 0x2433, 0x2435, 0x2437, 0x2439, 0x243b, 0x243d,
  0x243f, 0x2441, 0x2444, 0x2446, 0x2448, 0x244a, 0x244b, 0x244c,
  0x244d, 0x244e, 0x244f, 0x2452, 0x2455, 0x2458, 0x245b, 0x245e,
  0x245f, 0x2460, 0x2461, 0x2462, 0x2464, 0x2466, 0x2468, 0x2469,
  0x246a, 0x246b, 0x246c, 0x246d, 0x246f, 0x2473, 0x2470, 0x2471,
  0x246e, 0x242c, 0x242e, 0x2430, 0x2432, 0x2434, 0x2436, 0x2438,
  0x243a, 0x243c, 0x243e, 0x2440, 0x2442, 0x2445, 0x2447, 0x2449,
  0x2450, 0x2451, 0x2453, 0x2454, 0x2456, 0x2457, 0x2459, 0x245a,
  0x245c, 0x245d
};

static char *japan12Roman[10] = {
  "I", "II", "III", "IV", "V", "VI", "VII", "VIII", "IX", "X"
};

static char *japan12Abbrev1[6] = {
  "mm", "cm", "km", "mg", "kg", "cc"
};

#endif

extern double scale;
extern GBool mode;
extern char extension[5];
extern GBool ignore;
extern GBool printCommands;
extern GBool printHtml;
extern GBool noframes;
extern GBool stout;


//------------------------------------------------------------------------
// HtmlString
//------------------------------------------------------------------------





GString* basename(GString* str){
  
  char *p=str->getCString();
  int len=str->getLength();
  for (int i=len-1;i>=0;i--)
    if (*(p+i)=='/') 
      return new GString((p+i+1),len-i-1);
  return new GString(str);
}
     
GString* Dirname(GString* str){
  
  char *p=str->getCString();
  int len=str->getLength();
  for (int i=len-1;i>=0;i--)
    if (*(p+i)=='/') 
      return new GString(p,i+1);
  return new GString();
} 

HtmlString::HtmlString(GfxState *state, GBool hexCodes1,HtmlFontAccu* fonts) {
  double x, y, h;
  int f;
  state->transform(state->getCurX(), state->getCurY(), &x, &y);
  h = state->getTransformedFontSize();
  //~ yMin/yMax computation should use font ascent/descent values
  yMin = y - 0.95 * h;
  yMax = yMin + 1.3 * h;
  col = 0;
  text = new GString();
  if (state->getFont()->getName()) 
  fontpos=fonts->AddFont(HtmlFont(state->getFont()->getName(),static_cast<int>(h),state->getFillColor()));
  else  fontpos=0;
  xRight = NULL;
  yxNext = NULL;
  xyNext = NULL;
  hexCodes = hexCodes1;
  htext=new GString();
}

HtmlString::~HtmlString() {
  delete text;
  delete htext ;
  gfree(xRight);
}

void HtmlString::addChar(GfxState *state, double x, double y,
                         double dx, double dy,
                         Guchar c, GBool useASCII7) {
  char *charName, *sub;
  int c1;
  int i, j, n, m;

  // get current index
  i = text->getLength();

  // append translated character(s) to string
  sub = NULL;
  n = 1;
  if ((charName = state->getFont()->getCharName(c))) {
    if (useASCII7)
      c1 = ascii7Encoding.getCharCode(charName);
    else
      c1 = isoLatin1Encoding.getCharCode(charName);
    if (c1 < 0) {
      m = strlen(charName);
      if (hexCodes && m == 3 &&
          (charName[0] == 'B' || charName[0] == 'C' ||
           charName[0] == 'G') &&
          isxdigit(charName[1]) && isxdigit(charName[2])) {
        sscanf(charName+1, "%x", &c1);
      } else if (!hexCodes && m >= 2 && m <= 3 &&
                 isdigit(charName[0]) && isdigit(charName[1])) {
        c1 = atoi(charName);
        if (c1 >= 256)
          c1 = -1;
      } else if (!hexCodes && m >= 3 && m <= 5 && isdigit(charName[1])) {
        c1 = atoi(charName+1);
        if (c1 >= 256)
          c1 = -1;
      }
      //~ this is a kludge -- is there a standard internal encoding
      //~ used by all/most Type 1 fonts?
      if (c1 == 262)            // hyphen
        c1 = 45;
      else if (c1 == 266)       // emdash
        c1 = 208;
      if (useASCII7)
        c1 = ascii7Encoding.getCharCode(isoLatin1Encoding.getCharName(c1));
    }
    if (useASCII7) {
      if (c1 >= 128) {
        sub = ascii7Subst[c1 - 128];
        n = strlen(sub);
      }
    } else {
      if (c1 >= 256) {
        sub = isoLatin1Subst[c1 - 256];
        n = strlen(sub);
      }
    }
  } else {
    c1 = -1;
  }
  if (sub)
    text->append(sub);
  else if (c1 >= 0)
    text->append((char)c1);
  else
    text->append(' ');

  // update position information
  if (i+n > ((i+15) & ~15))
    xRight = (double *)grealloc(xRight, ((i+n+15) & ~15) * sizeof(double));
  if (i == 0)
    xMin = x;
  for (j = 0; j < n; ++j)
    xRight[i+j] = x + ((j+1) * dx) / n;
  xMax = x + dx;
}

void HtmlString::addChar16(GfxState *state, double x, double y,
                           double dx, double dy,
                           int c, GfxFontCharSet16 charSet) {
  int c1, t1, t2;
  int sub[8];
  char *p;
  int *q;
  int i, j, n;

  // get current index
  i = text->getLength();

  // convert the 16-bit character
  c1 = 0;
  sub[0] = 0;
  switch (charSet) {

  // convert Adobe-Japan1-2 to JIS X 0208-1983
  case font16AdobeJapan12:
#if JAPANESE_SUPPORT
    if (c <= 96) {
      c1 = 0x8080 + japan12Map[c];
    } else if (c <= 632) {
      if (c <= 230)
        c1 = 0;
      else if (c <= 324)
        c1 = 0x8080 + japan12Map[c - 230];
      else if (c <= 421)
        c1 = 0x8080 + japan12KanaMap1[c - 325];
      else if (c <= 500)
        c1 = 0;
      else if (c <= 598)
        c1 = 0x8080 + japan12KanaMap2[c - 501];
      else
        c1 = 0;
    } else if (c <= 1124) {
      if (c <= 779) {
        if (c <= 726)
          c1 = 0xa1a1 + (c - 633);
        else if (c <= 740)
          c1 = 0xa2a1 + (c - 727);
        else if (c <= 748)
          c1 = 0xa2ba + (c - 741);
        else if (c <= 755)
          c1 = 0xa2ca + (c - 749);
        else if (c <= 770)
          c1 = 0xa2dc + (c - 756);
        else if (c <= 778)
          c1 = 0xa2f2 + (c - 771);
        else
          c1 = 0xa2fe;
      } else if (c <= 841) {
        if (c <= 789)
          c1 = 0xa3b0 + (c - 780);
        else if (c <= 815)
          c1 = 0xa3c1 + (c - 790);
        else
          c1 = 0xa3e1 + (c - 816);
      } else if (c <= 1010) {
        if (c <= 924)
          c1 = 0xa4a1 + (c - 842);
        else
          c1 = 0xa5a1 + (c - 925);
      } else {
        if (c <= 1034)
          c1 = 0xa6a1 + (c - 1011);
        else if (c <= 1058)
          c1 = 0xa6c1 + (c - 1035);
        else if (c <= 1091)
          c1 = 0xa7a1 + (c - 1059);
        else
          c1 = 0xa7d1 + (c - 1092);
      }
    } else if (c <= 4089) {
      t1 = (c - 1125) / 94;
      t2 = (c - 1125) % 94;
      c1 = 0xb0a1 + (t1 << 8) + t2;
    } else if (c <= 7477) {
      t1 = (c - 4090) / 94;
      t2 = (c - 4090) % 94;
      c1 = 0xd0a1 + (t1 << 8) + t2;
    } else if (c <= 7554) {
      c1 = 0;
    } else if (c <= 7563) {     // circled Arabic numbers 1..9
      c1 = 0xa3b1 + (c - 7555);
    } else if (c <= 7574) {     // circled Arabic numbers 10..20
      t1 = c - 7564 + 10;
      sub[0] = 0xa3b0 + (t1 / 10);
      sub[1] = 0xa3b0 + (t1 % 10);
      sub[2] = 0;
      c1 = -1;
    } else if (c <= 7584) {     // Roman numbers I..X
      for (p = japan12Roman[c - 7575], q = sub; *p; ++p, ++q) {
        *q = 0xa380 + *p;
      }
      *q = 0;
      c1 = -1;
    } else if (c <= 7632) {
      if (c <= 7600) {
        c1 = 0;
      } else if (c <= 7606) {
        for (p = japan12Abbrev1[c - 7601], q = sub; *p; ++p, ++q) {
          *q = 0xa380 + *p;
        }
        *q = 0;
        c1 = -1;
      } else {
        c1 = 0;
      }
    } else {
      c1 = 0;
    }
#endif // JAPANESE_SUPPORT
    break;
  }

  // append converted character to string
  if (c1 == 0) {
#if 0 //~
    error(-1, "Unsupported Adobe-Japan1-2 character: %d", c);
#endif
    text->append(' ');
    n = 1;
  } else if (c1 > 0) {
    text->append(c1 >> 8);
    text->append(c1 & 0xff);
    n = 2;
  } else {
    n = 0;
    for (q = sub; *q; ++q) {
      text->append(*q >> 8);
      text->append(*q & 0xff);
      n += 2;
    }
  }

  // update position information
  if (i+n > ((i+15) & ~15)) {
    xRight = (double *)grealloc(xRight, ((i+n+15) & ~15) * sizeof(double));
  }
  if (i == 0) {
    xMin = x;
  }
  for (j = 0; j < n; ++j) {
    xRight[i+j] = x + dx;
  }
  xMax = x + dx;
}





//------------------------------------------------------------------------
// HtmlPage
//------------------------------------------------------------------------

HtmlPage::HtmlPage(GBool useASCII7, GBool rawOrder) {
  this->useASCII7 = useASCII7;
  this->rawOrder = rawOrder;
  curStr = NULL;
  yxStrings = NULL;
  xyStrings = NULL;
  yxCur1 = yxCur2 = NULL;
  images=new GVector<Image>();
  fonts=new HtmlFontAccu();
  links=new HtmlLinks();
  pageWidth=0;
  pageHeight=0; 
   
}

HtmlPage::~HtmlPage() {
  clear();
  delete images;
  delete fonts;
  delete links;
  
}

void HtmlPage::beginString(GfxState *state, GString *s, GBool hexCodes) {
  curStr = new HtmlString(state, hexCodes,fonts);
  if(!pageWidth) {
      pageWidth=static_cast<int>(state->getPageWidth());
      pageHeight=static_cast<int>(state->getPageHeight());
  }
}


void HtmlPage::conv(){
  HtmlString *tmp;
  GString* str;
  int k=0;
  for(tmp=yxStrings;tmp;tmp=tmp->yxNext){
     int pos=tmp->fontpos;
     //  printf("%d\n",pos);
     str=fonts->Get(pos).getFullName();
     if (tmp->htext) delete tmp->htext; 
     tmp->htext=HtmlFont::simple(str,tmp->text);
     delete str; 
   if (links->inLink(tmp->xMin,tmp->yMin,tmp->xMax,tmp->yMax,k))
         tmp->htext=links->getLink(k)->Link(tmp->htext);
  }
}


void HtmlPage::addChar(GfxState *state, double x, double y,
                       double dx, double dy, Guchar c) {
  double x1, y1, w1, h1, dx2, dy2;
  int n;
  GBool hexCodes;

  state->transform(x, y, &x1, &y1);
  state->textTransformDelta(state->getCharSpace(), 0, &dx2, &dy2);
  dx -= dx2;
  dy -= dy2;
  state->transformDelta(dx, dy, &w1, &h1);
  n = curStr->text->getLength();
  if (n > 0 &&
      x1 - curStr->xRight[n-1] > 0.1 * (curStr->yMax - curStr->yMin)) {
    hexCodes = curStr->hexCodes;
    endString();
    beginString(state, NULL, hexCodes);
  }
  curStr->addChar(state, x1, y1, w1, h1, c, useASCII7);
}

void HtmlPage::addChar16(GfxState *state, double x, double y,
                         double dx, double dy, int c,
                         GfxFontCharSet16 charSet) {
  double x1, y1, w1, h1, dx2, dy2;
  int n;
  GBool hexCodes;

  state->transform(x, y, &x1, &y1);
  state->textTransformDelta(state->getCharSpace(), 0, &dx2, &dy2);
  dx -= dx2;
  dy -= dy2;
  state->transformDelta(dx, dy, &w1, &h1);
  n = curStr->text->getLength();
  if (n > 0 &&
      x1 - curStr->xRight[n-1] > 0.1 * (curStr->yMax - curStr->yMin)) {
    hexCodes = curStr->hexCodes;
    endString();
    beginString(state, NULL, hexCodes);
  }
  curStr->addChar16(state, x1, y1, w1, h1, c, charSet);
}

void HtmlPage::endString() {
  HtmlString *p1, *p2;
  double h, y1, y2;

  // throw away zero-length strings -- they don't have valid xMin/xMax
  // values, and they're useless anyway
  if (curStr->text->getLength() == 0) {
    delete curStr;
    curStr = NULL;
    return;
  }

#if 0 //~tmp
  if (curStr->yMax - curStr->yMin > 20) {
    delete curStr;
    curStr = NULL;
    return;
  }
#endif

  // insert string in y-major list
  h = curStr->yMax - curStr->yMin;
  y1 = curStr->yMin + 0.5 * h;
  y2 = curStr->yMin + 0.8 * h;
  if (rawOrder) {
    p1 = yxCur1;
    p2 = NULL;
  } else if ((!yxCur1 ||
              (y1 >= yxCur1->yMin &&
               (y2 >= yxCur1->yMax || curStr->xMax >= yxCur1->xMin))) &&
             (!yxCur2 ||
              (y1 < yxCur2->yMin ||
               (y2 < yxCur2->yMax && curStr->xMax < yxCur2->xMin)))) {
    p1 = yxCur1;
    p2 = yxCur2;
  } else {
    for (p1 = NULL, p2 = yxStrings; p2; p1 = p2, p2 = p2->yxNext) {
      if (y1 < p2->yMin || (y2 < p2->yMax && curStr->xMax < p2->xMin))
        break;
    }
    yxCur2 = p2;
  }
  yxCur1 = curStr;
  if (p1)
    p1->yxNext = curStr;
  else
    yxStrings = curStr;
  curStr->yxNext = p2;
  curStr = NULL;
}

void HtmlPage::coalesce() {
  HtmlString *str1, *str2;
  double space, d;
  int n, i;

#if 0 //~ for debugging
  for (str1 = yxStrings; str1; str1 = str1->yxNext) {
    printf("x=%3d..%3d  y=%3d..%3d  size=%2d '%s'\n",
           (int)str1->xMin, (int)str1->xMax, (int)str1->yMin, (int)str1->yMax,
           (int)(str1->yMax - str1->yMin), str1->text->getCString());
  }
  printf("\n------------------------------------------------------------\n\n");
#endif
  str1 = yxStrings;
  while (str1 && (str2 = str1->yxNext)) {
    space = str1->yMax - str1->yMin;
    d = str2->xMin - str1->xMax;
#if 0 //~tmp
    if (((rawOrder &&
          ((str2->yMin >= str1->yMin && str2->yMin <= str1->yMax) ||
           (str2->yMax >= str1->yMin && str2->yMax <= str1->yMax))) ||
         (!rawOrder && str2->yMin < str1->yMax)) &&
        d > -0.1 * space && d < 0.2 * space &&(str1->fontpos==str2->fontpos)){
#else
    if (((rawOrder &&
          ((str2->yMin >= str1->yMin && str2->yMin <= str1->yMax) ||
           (str2->yMax >= str1->yMin && str2->yMax <= str1->yMax))) ||
         (!rawOrder && str2->yMin < str1->yMax)) &&
        d > -0.5 * space && d < space&&(str1->fontpos==str2->fontpos)) {
#endif
      n = str1->text->getLength();
      
            
      if (d > 0.1 * space){
	str1->text->append(" ");
        str1->htext->append(" \n");
      }
      
        str1->text->append(str2->text);
        str1->htext->append(str2->htext);
        str1->xRight = (double *)
	grealloc(str1->xRight, str1->text->getLength() * sizeof(double));
      if (d > 0.1 * space)
	str1->xRight[n++] = str2->xMin;
      for (i = 0; i < str2->text->getLength(); ++i)
	str1->xRight[n++] = str2->xRight[i];
      if (str2->xMax > str1->xMax)
	str1->xMax = str2->xMax;
      if (str2->yMax > str1->yMax)
	str1->yMax = str2->yMax;
      str1->yxNext = str2->yxNext;
      delete str2;
      
    } else {
      str1 = str2;
    }
          
 }
}



  



       
void HtmlPage::AddImage(const Image& x){
  images->push_back(x);
}

void HtmlPage::dump(FILE *f) {
  if (mode){
  
     fputs("<style type=\"text/css\">\n<!--\n",f);
     fputs(".pg{position:absolute;top:0px;left:0px;height:",f);
     fputs(GString::IntToStr(xoutRound((double) pageHeight * scale*1.1))->getCString(),f);
     fputs(";width:",f);
     fputs(GString::IntToStr(xoutRound((double)pageWidth*scale*1.1))->getCString(),f);
     fputs(";}\n\n",f);
     for(int i=0;i!=fonts->size();i++){
         fputs(fonts->CSStyle(i)->getCString(),f);
         fputc('\n',f);
     }
  
  fputs("-->\n</style>\n",f);
  fputs("</head>\n<body bgcolor=#FFFFFF vlink=blue link=blue>\n",f);  
  
  fputs("<div class=\"pg\"></div>\n",f);
  }
  
  for(GVector<Image>::iterator i=images->begin();i!=images->end();i++){
    GString *tmp1=i->getHtmlString();
    int k=0;
     if (links->inLink((double)i->getX(),(double)i->getY(),(double)(i->getX()+i->getWidth()),
                       (double)(i->getY()+i->getHeight()),k))
         tmp1=links->getLink(k)->Link(tmp1);
     if (mode){ 
       fputs("<layer top=",f);
       fputs(GString::IntToStr(xoutRound((i->getY())*(scale)))->getCString(),f);
       fputs(" left=",f);
       fputs(GString::IntToStr(xoutRound((i->getX())*(scale)))->getCString(),f);
       fputs(">",f);
     }
    fputs(tmp1->getCString(),f);
     if (mode)  fputs("</layer>\n",f);
      else fputs("<br>\n",f); 
    delete tmp1;
  }

  GString* str;
  for(HtmlString *tmp=yxStrings;tmp;tmp=tmp->yxNext){
   if (tmp->htext){
    str=new GString(tmp->htext);
    if(mode){
       fputs("<layer top=",f);
       fputs(GString::IntToStr(xoutRound((tmp->yMin)*(scale)))->getCString(),f);
       fputs(" left=",f);
       fputs(GString::IntToStr(xoutRound((tmp->xMin)*(scale)))->getCString(),f);
       fputs(">",f); 
       if (tmp->fontpos!=-1){
          str=fonts->getCSStyle(tmp->fontpos,str);
       }
    }    
    fputs(str->getCString(),f);
    delete str;      
    if (mode)  fputs("</layer>\n",f); 
    else 
      fputs("<br>\n",f);
   }
  }
 
 if (mode) fputs("</body>\n</html>\n",f);
 
}



void HtmlPage::clear() {
  HtmlString *p1, *p2;

  if (curStr) {
    delete curStr;
    curStr = NULL;
  }
  for (p1 = yxStrings; p1; p1 = p2) {
    p2 = p1->yxNext;
     delete p1;
  }
  yxStrings = NULL;
  xyStrings = NULL;
  yxCur1 = yxCur2 = NULL;
  delete images;
  delete fonts;
  delete links;
 
  images=new GVector<Image>();
  fonts=new HtmlFontAccu();
  links=new HtmlLinks();
 

}

//------------------------------------------------------------------------
// HtmlOutputDev
//------------------------------------------------------------------------



HtmlOutputDev::HtmlOutputDev(char *fileName, GBool useASCII7, GBool rawOrder) {
  f=NULL;
  FILE *ind;
  pages = NULL;
  dumpJPEG=gTrue;
  this->rawOrder = rawOrder;
  ok = gTrue;
  imgNum=1;
  pageNum=1;
  // open file
  needClose = gFalse;
  HtmlFont::setScale(scale);
  Image::setScale(scale); 
  Docname=new GString (fileName);
  GString *tmp=basename(Docname);
  if (tmp->getLength()==0) {printf("Error : illegal output file");exit(0);} 
  GString *tmp1=new GString(fileName);
  tmp1->append(".html");
  GString *tmp2=new GString(Docname);
  tmp2->append("_ind.html"); 
  
  if (!(!mode&&noframes)){ 
    if (fileName) {
      if (!strcmp(fileName, "-")) {
        ind = stdout;
      } else if (!(ind = fopen(tmp1->getCString(), "w"))) 
        {
        error(-1, "Couldn't open html file '%s'", tmp1->getCString());
        return;
      }
    } 
    if (ind) {
      fputs("<html>\n<head>\n<title>\n</title>\n<frameset border=3 frameborder=\"yes\" framespacing=2 cols=\" 130,* \">",ind);
      fputs("\n<frame name=\"links\" src=\"",ind);
      fputs(tmp->getCString(),ind);
      fputs("_ind.html",ind);
      fputs("\" target=\"rechts\">\n<frame name=\"rechts\"\n src=\"",ind); 
      if (mode) {
          tmp->append("-1.html");
          fputs(tmp->getCString(),ind);
       }
      else{ 
      tmp->append("s.html");
      fputs(tmp->getCString(),ind);
    }
     fputs("\">\n</frameset>\n</html>\n",ind);
   }
    fclose(ind); 


    if ((f = fopen(tmp2->getCString(), "w"))) {
      needClose = gTrue;
    } else {
      error(-1, "Couldn't open html file '%s'", tmp2->getCString());
      ok = gFalse;
      return;
    }  
   fputs("<html>\n<head>\n<title></title>\n</head>\n<body bgcolor=#A0A0A0 vlink=blue link=blue>",f);  
  }
    
    if ((tin = fopen(Dirname(Docname)->append("images.log")->getCString(), "w"))) {
      needClose = gTrue;
    } else {
      error(-1, "Couldn't open file '%s'", "images.log");
      ok = gFalse;
      return;
    }  

 
  
  // set up text object
  pages = new HtmlPage(useASCII7, rawOrder);
  delete tmp;
  delete tmp1;
  delete tmp2;
}

HtmlOutputDev::~HtmlOutputDev() {
 delete Docname;
 if (needClose){
   if (f){
    fputs("</body></html>\n",f);  
    fclose(f);
   }
}
 if (!mode){ 
    fputs("</body>\n</html>\n",page);  
    fclose(page);
 }
  if (pages)
    delete pages;
    
  if (tin)
     fclose(tin);
}



void HtmlOutputDev::startPage(int pageNum, GfxState *state) {
 GString *str=basename(Docname);
  pages->clear();
  GString *tmp=new GString(Docname); 
  char *fName;  
  GString *pgNum=GString::IntToStr(pageNum);
  if(!(noframes&&!mode)){
   if (f){
    if (mode){
      fputs("<a href=\"",f);
      fputs(str->getCString(),f);
      fputs("-",f);
	}
     else{ 
      fputs("<a href=\"",f);
      fputs(str->getCString(),f);
      fputs("s.html#",f);
     }
    fputs(pgNum->getCString(),f);
    if (mode)
        fputs(".html\"",f);
      else
        fputs("\"",f);
    fputs(" target=\"rechts\" >Page ",f);
    fputs(pgNum->getCString(),f);
    fputs("</a>\n",f);
   }
  } 
    if (mode){
      tmp->append("-");
      tmp->append(pgNum);
    } 
     else
	if (pageNum==1) 
           if(!noframes){tmp->append("s");}
    tmp->append(".html");
      fName=tmp->getCString();
      if(stout) page=stdout;
      else 
      if(mode||(pageNum==1)){
        if (!(page=fopen(fName,"wb"))) {
         error(-1, "Couldn't open HTMLfile '%s'", fName);
         return;
       }
      }   
      delete str;
      delete tmp;
    if(mode || (pageNum==1)){
    fputs("<html>\n",page);
    fputs("<head>\n",page);
    fputs("<title></title>\n",page);
    }
    if (!mode&&(pageNum==1))  fputs("</head>\n<body bgcolor=#FFFFFF vlink=blue link=blue>\n",page);
   if (page){ 
    if (!mode){
        fputs("<br>\n<a name=",page);
        fputs(pgNum->getCString(),page);
        fputs("><b>Page ",page);
        fputs(pgNum->getCString(),page);
        fputs("</b></a><br><br>",page);
     }
   }
   
if (pgNum) delete pgNum;
}


void HtmlOutputDev::endPage() {
  pageNum++;
  pages->conv();
  pages->coalesce();
    pages->dump(page);
   if (mode) fclose(page);  
  if(!noframes) fputs("<BR>", f);
   printf("Page-%d\n",(pageNum-1));
 
}

void HtmlOutputDev::updateFont(GfxState *state) {
  GfxFont *font;
  char *charName;
  int c;

  // look for hex char codes in subsetted font
  hexCodes = gFalse;
  if ((font = state->getFont()) && !font->is16Bit()) {
    for (c = 0; c < 256; ++c) {
      if ((charName = font->getCharName(c))) {
	if ((charName[0] == 'B' || charName[0] == 'C' ||
	     charName[0] == 'G') &&
	    strlen(charName) == 3 &&
	    ((charName[1] >= 'a' && charName[1] <= 'f') ||
	     (charName[1] >= 'A' && charName[1] <= 'F') ||
	     (charName[2] >= 'a' && charName[2] <= 'f') ||
	     (charName[2] >= 'A' && charName[2] <= 'F'))) {
	  hexCodes = gTrue;
	  break;
	}
      }
    }
  }
}

void HtmlOutputDev::beginString(GfxState *state, GString *s) {
  pages->beginString(state, s, hexCodes);
}

void HtmlOutputDev::endString(GfxState *state) {
  pages->endString();
}

void HtmlOutputDev::drawChar(GfxState *state, double x, double y,
			     double dx, double dy, Guchar c) {
  pages->addChar(state, x, y, dx, dy, c);
}

void HtmlOutputDev::drawChar16(GfxState *state, double x, double y,
			       double dx, double dy, int c) {
  pages->addChar16(state, x, y, dx, dy, c, state->getFont()->getCharSet16());
}


void HtmlOutputDev::drawImageMask(GfxState *state, Stream *str,
				   int width, int height, GBool invert,
				   GBool inlineImg) {
if (ignore) return;
  FILE *f1;
  int c;
  
  int x0, y0;			// top left corner of image
  int w0, h0, w1, h1;		// size of image
  double xt, yt, wt, ht;
  GBool rotate, xFlip, yFlip;
  GBool dither;
  int x, y;
  int ix, iy;
  int px1, px2, qx, dx;
  int py1, py2, qy, dy;
  Gulong pixel;
  int nComps, nVals, nBits;
  double r1, g1, b1;
  int i, j;

  // get image position and size
  state->transform(0, 0, &xt, &yt);
  state->transformDelta(1, 1, &wt, &ht);
  if (wt > 0) {
    x0 = xoutRound(xt);
    w0 = xoutRound(wt);
  } else {
    x0 = xoutRound(xt + wt);
    w0 = xoutRound(-wt);
  }
  if (ht > 0) {
    y0 = xoutRound(yt);
    h0 = xoutRound(ht);
  } else {
    y0 = xoutRound(yt + ht);
    h0 = xoutRound(-ht);
  }
  state->transformDelta(1, 0, &xt, &yt);
  rotate = fabs(xt) < fabs(yt);
  if (rotate) {
    w1 = h0;
    h1 = w0;
    xFlip = ht < 0;
    yFlip = wt > 0;
  } else {
    w1 = w0;
    h1 = h0;
    xFlip = wt < 0;
    yFlip = ht > 0;
  }

  GString *fName1=basename(Docname);
  GString *fName=new GString(Docname);
  fName->append("-");
  fName1->append("-");
  Image* image=NULL;
  GString *pgNum=GString::IntToStr(pageNum);
  GString *imgnum=GString::IntToStr(imgNum);

  // dump JPEG file
  if (dumpJPEG  && str->getKind() == strDCT) {
    
    // open the image file
    fName->append(pgNum)->append("_")->append(imgnum)->append(".jpg");
    fName1->append(pgNum)->append("_")->append(imgnum)->append(".jpg");
    ++imgNum;
    if (!(f1 = fopen(fName->getCString(), "wb"))) {
      error(-1, "Couldn't open image file '%s'", fName->getCString());
      return;
    }

    // initialize stream
    str = ((DCTStream *)str)->getRawStream();
    str->reset();

    // copy the stream
    while ((c = str->getChar()) != EOF)
      fputc(c, f1);

    fclose(f1);
   

    image=new Image(fName1,w0,h0,x0,y0);
    pages->AddImage(*image);
    
  } 
   else
#ifndef DB_ALT
     if (mode) 
#else
     if (1) 
#endif
    {
    fName1->append(pgNum)->append("_")->append(imgnum);
    fName1->append(extension);
    fName->append(pgNum)->append("_")->append(imgnum)->append(".pbm");
    ++imgNum;
   
    fputs(basename(fName)->getCString(),tin);
    fputc('\n',tin);
     
    if (!(f1 = fopen(fName->getCString(), "wb"))) {
      error(-1, "Couldn't open image file '%s'", fName->getCString());
      return;
       }
    fprintf(f1, "P4\n");
    fprintf(f1, "%d %d\n", width, height);

    // initialize stream
    str->reset();

    // copy the stream
    while ((c = str->getChar()) != EOF)
      fputc(c, f1);

    fclose(f1);

    image=new Image(fName1,w0,h0,x0,y0);
    pages->AddImage(*image);
    
    }
  
  if (image) delete image;
   delete pgNum;
   delete imgnum;
   delete fName1;
   delete fName;
}

void HtmlOutputDev::drawImage(GfxState *state, Stream *str, int width,
			       int height, GfxImageColorMap *colorMap,
			       GBool inlineImg) {
  

if (ignore) return;
  FILE *f1;
  ImageStream *imgStr;
  Guchar pixBuf[4];
  GfxColor color;
  int c;
  
  int x0, y0;			// top left corner of image
  int w0, h0, w1, h1;		// size of image
  double xt, yt, wt, ht;
  GBool rotate, xFlip, yFlip;
  GBool dither;
  int x, y;
  int ix, iy;
  int px1, px2, qx, dx;
  int py1, py2, qy, dy;
  Gulong pixel;
  int nComps, nVals, nBits;
  double r1, g1, b1;
  int i, j;

  // get image position and size
  state->transform(0, 0, &xt, &yt);
  state->transformDelta(1, 1, &wt, &ht);
  if (wt > 0) {
    x0 = xoutRound(xt);
    w0 = xoutRound(wt);
  } else {
    x0 = xoutRound(xt + wt);
    w0 = xoutRound(-wt);
  }
  if (ht > 0) {
    y0 = xoutRound(yt);
    h0 = xoutRound(ht);
  } else {
    y0 = xoutRound(yt + ht);
    h0 = xoutRound(-ht);
  }
  state->transformDelta(1, 0, &xt, &yt);
  rotate = fabs(xt) < fabs(yt);
  if (rotate) {
    w1 = h0;
    h1 = w0;
    xFlip = ht < 0;
    yFlip = wt > 0;
  } else {
    w1 = w0;
    h1 = h0;
    xFlip = wt < 0;
    yFlip = ht > 0;
  }

  GString *fName1=basename(Docname);
  GString *fName=new GString(Docname);
  fName->append("-");
  fName1->append("-");
  GString *pgNum= GString::IntToStr(pageNum);
  GString *imgnum= GString::IntToStr(imgNum);  
  Image *image=NULL;

  // dump JPEG file
  if (dumpJPEG && str->getKind() == strDCT) {
    
    // open the image file
    fName->append(pgNum)->append("_")->append(imgnum)->append(".jpg");
    fName1->append(pgNum)->append("_")->append(imgnum)->append(".jpg");
    ++imgNum;
    
    if (!(f1 = fopen(fName->getCString(), "wb"))) {
      error(-1, "Couldn't open image file '%s'", fName->getCString());
      return;
    }

    // initialize stream
    str = ((DCTStream *)str)->getRawStream();
    str->reset();

    // copy the stream
    while ((c = str->getChar()) != EOF)
      fputc(c, f1);
    
    
    image=new Image(fName1,w0,h0,x0,y0);
    pages->AddImage(*image);
    
   
    fclose(f1);
  } else 
#ifndef DB_ALT
    if (mode) 
#else
    if (1) 
#endif
     {

     // open the image file and write the PPM header
    fName1->append(pgNum)->append("_")->append(imgnum)->append(extension);
    fName->append(pgNum)->append("_")->append(imgnum)->append(".ppm");

    fputs(basename(fName)->getCString(),tin);
    fputc('\n',tin);

    ++imgNum;
    if (!(f1 = fopen(fName->getCString(), "wb"))) {
      error(-1, "Couldn't open image file '%s'", fName->getCString());
      return;
    }
    fprintf(f1, "P6\n");
    fprintf(f1, "%d %d\n", width, height);
    fprintf(f1, "255\n");

    // initialize stream
    imgStr = new ImageStream(str, width, colorMap->getNumPixelComps(),
			     colorMap->getBits());
    imgStr->reset();

    // for each line...
    for (y = 0; y < height; ++y) {

      // write the line
      for (x = 0; x < width; ++x) {
	imgStr->getPixel(pixBuf);
	colorMap->getColor(pixBuf, &color);
	fputc((int)(color.getR() * 255 + 0.5), f1);
	fputc((int)(color.getG() * 255 + 0.5), f1);
	fputc((int)(color.getB() * 255 + 0.5), f1);
      }
    }
   
    
    image=new Image(fName1,w0,h0,x0,y0);
    pages->AddImage(*image);
   
    delete imgStr;

    fclose(f1);
   
  }
   if(image) delete image;
    delete fName1;
    delete fName;
    delete pgNum;
    delete imgnum;
}

void HtmlOutputDev::drawLink(Link* link,Catalog *cat){
  double _x1,_y1,_x2,_y2,w;
  int x1,y1,x2,y2;

  link->getBorder(&_x1,&_y1,&_x2,&_y2,&w);
  cvtUserToDev(_x1,_y1,&x1,&y1);
  cvtUserToDev(_x2,_y2,&x2,&y2); 
  GString* _dest=getLinkDest(link,cat);
  HtmlLink t((double) x1,(double) y2,(double) x2,(double) y1,_dest);
  pages->AddLink(t);
  delete _dest;
}

GString *HtmlOutputDev::getLinkDest(Link *link,Catalog* catalog){
  char *p;
  switch(link->getAction()->getKind()) {
      case actionGoTo:{ 
	          GString* file=basename(Docname);
                  int page=1;
                  LinkGoTo *ha=(LinkGoTo *)link->getAction();
                  LinkDest *dest=NULL;
                  if (ha->getDest()==NULL) dest=catalog->findDest(ha->getNamedDest());
                  else dest=ha->getDest();
                  if (dest){ 
                    if (dest->isPageRef()){
                        Ref pageref=dest->getPageRef();
                        page=catalog->findPage(pageref.num,pageref.gen);
		   }
                    else  page=dest->getPageNum();
                  
                   GString *str=GString::IntToStr(page);
                   if (mode) file->append("-");
		   else{ 
                       if (!noframes) file->append("s");
		       file->append(".html#");
                    }
                   file->append(str);
                   if (mode) file->append(".html");
                  if (printCommands) printf(" page %d ",page);
                  delete str;
                  return file;
                  }
                  else return new GString();
      }
       
      case actionGoToR:{
                  
                  LinkGoToR *ha=(LinkGoToR *) link->getAction();
                  LinkDest *dest=NULL;
                  int page=1;
                  GString *file=new GString();
                  if (ha->getFileName()){
                    delete file;
                    file=new GString(ha->getFileName()->getCString());
                  }
                  if (ha->getDest()!=NULL)  dest=ha->getDest();
		  if (dest&&file){
                  if (!(dest->isPageRef()))  page=dest->getPageNum();
                  if (printCommands) printf(" page %d ",page);
		   if (printHtml){
                     p=file->getCString()+file->getLength()-4;
		     if (!strcmp(p, ".pdf") || !strcmp(p, ".PDF")){
                      file->del(file->getLength()-4,4);
                      file->append(".html");
		     }
                    file->append('#');
                    file->append(GString::IntToStr(page));
		   }
                  }
                  if (printCommands) printf("filename %s\n",file->getCString());
                  return file;
      
           
 
      }
      case actionURI: {
                  LinkURI *ha=(LinkURI *) link->getAction();
                  GString* file=new GString(ha->getURI()->getCString());
                  // printf("uri : %s\n",file->getCString());
                  return file;
  }
       
      case actionLaunch: {
                  LinkLaunch *ha=(LinkLaunch *) link->getAction();
                  GString* file=new GString(ha->getFileName()->getCString());
		   if (printHtml) { 
                      p=file->getCString()+file->getLength()-4;
		     if (!strcmp(p, ".pdf") || !strcmp(p, ".PDF")){
                        file->del(file->getLength()-4,4);
                        file->append(".html");
		     }
                    if (printCommands) printf("filename %s",file->getCString());
		     }
                  return file;      
     }
}
}









